--[[
==============================================================================
Trade Assistant 2.0 - Plugin for Vendetta Online (all platforms)
==============================================================================
Created by Mad Miner Moda in 2014, fixed and improved by Snib in 2022.

Store sellable item data relevant for trading (prices, cu and kg) and provide
a GUI for finding profitable trade routes and where item are sold.

Visit each single station in game to record their inventory. Note that some 
stations are fully dynamic (currently only Latos stations) and their inventory 
changes depending on what gets sold there.

Prices stations buy at depend on player supply, whereas prices they sell at are 
usually static, although commodity prices are modified by the player's trade 
skill level and standing with the faction operating the station.

Sell prices can only be obtained for items you have in your inventory (in your 
ship's cargohold or stored at the station you are in).

Commands:
   /ta          -- show GUI
   /ta version  -- print Trade Assist version
   /ta purge    -- reset database, all existing data will be lost
==============================================================================
]]

-- set to false if ship protoype mass bug ever gets fixed, see
-- https://www.vendetta-online.com/x/msgboard/2/38113
local shipMassBugWorkaround = true

local plugin_path = "plugins/TradeAssist/"
local notesid = 20100201
local command = "ta"

local floor = math.floor
local fmod = math.fmod
local sub = string.sub
local format = string.format
local sorted = {}
local timer = Timer()
local init_done = false
local is_premium = false
local activeShipMass = 0
local has_old_goods = false

local TA
TA = {
    stations = {}, -- list of stations and related info here.
    goods = {}, -- trade goods table. this is where the magic happens.
    goods_old = {}, -- old format trade goods table gets imported to this
    Version = "2.0~Snib",
    GUI = {
        Color = { -- also controls the sort order for the goods tab
            WHITE        = rgbtohex("255 255 255"),
            GREEN        = rgbtohex("51 204 51"),
            ORANGE       = rgbtohex("255 165 0"),
            commodities  = rgbtohex("55 210 255"),
            ship         = rgbtohex("126 255 128"),
            ore          = rgbtohex("180 180 180"),
            mining       = rgbtohex("210 115 185"),
            lightweapon  = rgbtohex("218 110 75"),
            heavyweapon  = rgbtohex("220 40 80"),
            turretweapon = rgbtohex("254 69 0"),
            battery      = rgbtohex("255 185 53"),
        },
        Factioncolors = {
            Gray = rgbtohex(FactionColor_RGB[0]),
            Itani = rgbtohex(FactionColor_RGB[1]),
            Serco = rgbtohex(FactionColor_RGB[2]),
            UIT = rgbtohex(FactionColor_RGB[3]),
            Developers = rgbtohex("255 0 255")
        },
        Systemownership = {
            "Serco", -- 1: Sol II
            "Serco", -- 2: Betheshee
            "Serco", -- 3: Geira Rutilus
            "Itani", -- 4: Deneb
            "Itani", -- 5: Eo
            "Itani", -- 6: Cantus
            "Itani", -- 7: Metana
            "Itani", -- 8: Setalli Shinas
            "Itani", -- 9: Itan
            "Itani", -- 10: Pherona
            "Itani", -- 11: Artana Aquilus
            "Itani", -- 12: Divinia
            "Itani", -- 13: Jallik
            "Gray", -- 14: Edras
            "UIT", -- 15: Verasi
            "Gray", -- 16: Pelatus
            "Gray", -- 17: Bractus
            "UIT", -- 18: Nyrius
            "UIT", -- 19: Dau
            "Gray", -- 20: Sedina
            "UIT", -- 21: Azek
            "Gray", -- 22: Odia
            "Gray", -- 23: Latos
            "UIT", -- 24: Arta Caelestis
            "Gray", -- 25: Ukari
            "Gray", -- 26: Helios
            "Serco", -- 27: Initros
            "Serco", -- 28: Pyronis
            "Serco", -- 29: Rhamus
            "Serco", -- 30: Dantia
            "Developers", -- 31: Devlopia
            "Developers", -- 32: (Qualcomm Demo system)
            "Gray" -- 33: Unknown System
        },
        Hide = function() return iup.CLOSE end,
    },
    Print = function(str)
        print(TA.GUI.Color.WHITE .. "[" .. TA.GUI.Color.ORANGE .. "TA" ..
            TA.GUI.Color.WHITE .. "] " .. tostring(str))
    end,
}
TA.Title = "Trade Assistant v" .. TA.Version

local function save_data_now()
    timer:Kill()
    SaveSystemNotes(spickle({ version = TA.Version, goods = TA.goods,
        stations = TA.stations, goods_old = TA.goods_old }), notesid)
end

local function save_data()
    timer:SetTimeout(90000, save_data_now)
end

local function create_matrix(columndefs, defsort)
    local matrix
    local sortedlist
    local sort_key = defsort -- default to sort by distance
    local numcolumns = (#columndefs)
    local function set_sort_mode(mode)
        -- clicked on title of column
        sort_key = mode
        -- color the text accordingly
        for i = 1, numcolumns do
            matrix:setattribute("bgcolor", 0, i, mode == i and
                tabseltextcolor or tabunseltextcolor)
        end
    end

    local function sort_list(key, list)
        table.sort(list, columndefs[key].fn)
    end

    local function reload_matrix(list)
        matrix.numlin = (#list)

        for index, v in ipairs(list) do
            columndefs.update_entry(matrix, index, v)
        end
    end

    local update_matrix = function()
        set_sort_mode(sort_key)
        sort_list(sort_key, sortedlist)
        reload_matrix(sortedlist)
    end

    matrix = iup.pdasubsubmatrix {
        numcol = numcolumns,
        expand = "YES",
        click_cb = function(_, row, col)
            if row == 0 and tonumber(matrix.NUMLIN) > 0 then
                set_sort_mode(col)
                update_matrix()
            elseif columndefs.on_sel and sortedlist then
                columndefs.on_sel(matrix, row, sortedlist[row])
                update_matrix()
            end
        end,
        edition_cb = function()
            return iup.IGNORE
        end,
        bgcolor_cb = function(_, row)
            local c = ListColors.Numbers[fmod(row, 2)]
            return c[1], c[2], c[3], c[4], iup.DEFAULT
        end,
        update = update_matrix,
        reset = function(_, characterlist)
            matrix.dellin = "1--1"
            sortedlist = characterlist
            sort_list(sort_key, sortedlist)
            reload_matrix(sortedlist)
        end,
    }
    for i = 1, numcolumns do
        matrix["ALIGNMENT" .. i] = columndefs[i].alignment
        matrix["0:" .. i] = columndefs[i].title
    end
    set_sort_mode(defsort)

    return matrix
end

local function pad(d) return ("%09d%s"):format(#d, d) end

local function sortfunc(a, b)
    return tostring(a):gsub("%d+", pad) < tostring(b):gsub("%d+", pad)
end

-- returns a hex color for a name/type combination
local function get_goods_color(name, type)
    local color
    if name:match(" Ore") then
        color = TA.GUI.Color["ore"]
    elseif name:match("Mining") or name:match("Mineral Scanner") then
        color = TA.GUI.Color["mining"]
    elseif name:match("Turret") then
        color = TA.GUI.Color["turretweapon"]
    elseif TA.GUI.Color[type] then
        color = TA.GUI.Color[type]
    else
        color = TA.GUI.Color.WHITE
    end
    return color
end

local function location_string(stationid)
    local station = TA.stations[stationid]
    local class = GetStationClass(stationid)
    if class == 0 then class = "Conquerable" end
    return TA.GUI.Factioncolors[TA.GUI.Systemownership[station.systemid]] ..
        ShortLocationStr(station.sectorid) .. "\127o [" ..
        FactionName[GetStationFaction(stationid)] .. " " ..
        class .. "]" .. (station.dynamic and
            (TA.GUI.Color.ORANGE .. " D\127o") or "")
end

local function price_string(credits, crystals)
    credits = tonumber(credits)
    crystals = tonumber(crystals)
    if credits and credits > 0 then
        if is_premium then crystals = nil end
        return CreditsPlusCrystal(credits, crystals)
    end
    return ""
end

-- this extracts the credits part of a formatted price string for easy comparison
local function price_string_to_number(price)
    --price = price ~= "" and price:gsub(",", ""):gsub("+",""):gsub(CrystalIcon, ""):gsub(" ", "") or nil
    price = price:gsub(",", ""):match("%d+")
    return tonumber(price)
end

TA.GUI.goods_cu = iup.label { title = "", expand = "HORIZONTAL", alignment = "ARIGHT" }
TA.GUI.goods_mass = iup.label { title = "", expand = "HORIZONTAL", alignment = "ARIGHT" }
TA.GUI.goods_list_typeids = {}
TA.GUI.goods_list = iup.stationsublist { dropdown = "YES", expand = "HORIZONTAL",
    action = function(_, _, index, state)
        if state == 0 then return end
        local typeid = TA.GUI.goods_list_typeids[index]
        TA.GUI.goods_cu.title = comma_value(TA.goods[typeid].vol)
        TA.GUI.goods_mass.title = comma_value(TA.goods[typeid].mass)
        TA.GUI.goods_tbl = {}
        for k, v in pairs(TA.goods[typeid]) do
            if type(v) == "table" then
                table.insert(TA.GUI.goods_tbl, 1, {
                    name = location_string(k),
                    sell = price_string(v.s, v.s2),
                    buy = price_string(v.b, v.v2),
                }
                )
            end
        end
        TA.GUI.goods_matrix:reset(TA.GUI.goods_tbl)
    end
}
local clicked = {}
local click_timer = Timer()
local click_timeout = function()
    clicked = {}
end

local goods_sorttbl = {
    [1] = {
        title = "Name",
        alignment = "ALEFT",
        fn = function(a, b) return a.name < b.name end,
    },
    [2] = {
        title = "Buy",
        alignment = "ARIGHT",
        fn = function(a, b)
            if a.buy == b.buy then return a.name < b.name
            else return (price_string_to_number(a.buy) or 0) >
                    (price_string_to_number(b.buy) or 0)
            end
        end,
    },
    [3] = {
        title = "Sell",
        alignment = "ARIGHT",
        fn = function(a, b)
            if a.sell == b.sell then return a.name < b.name
            else return (price_string_to_number(a.sell) or 9999999999) <
                    (price_string_to_number(b.sell) or 9999999999)
            end
        end,
    },
    update_entry = function(self, index, data)
        self:setcell(index, 1, " " .. data.name)
        self:setcell(index, 2, " " .. data.buy)
        self:setcell(index, 3, " " .. data.sell)
    end,
    on_sel = function(_, row, data)
        if clicked[row] then
            local location_str = filter_colorcodes(data.name):match("(.-)%[")
            if location_str then
                local sectorid = SectorIDFromLocationStr(location_str)
                if sectorid and not IsTrainingSector(sectorid) then
                    NavRoute.SetFinalDestination(sectorid)
                    TA.Print("Navigation destination set to " .. location_str)
                end
            end
            clicked = {}
            click_timer:Kill()
        else
            clicked = {}
            clicked[row] = true
            click_timer:SetTimeout(500, click_timeout)
        end
    end,
}
TA.GUI.goods_matrix = create_matrix(goods_sorttbl, 1)
local goods_tab = iup.vbox {
    iup.hbox {
        iup.fill { size = 4 },
        iup.label { title = "Item: " }, TA.GUI.goods_list,
        iup.fill { size = 5 },
        iup.label { title = "Volume: " }, TA.GUI.goods_cu,
        iup.label { title = " cu", expand = "HORIZONTAL", alignment = "ALEFT" },
        iup.fill { size = 8 },
        iup.label { title = "Mass: " }, TA.GUI.goods_mass,
        iup.label { title = " kg", expand = "HORIZONTAL", alignment = "ALEFT" },
        iup.fill { size = 4 },
    },
    TA.GUI.goods_matrix,
    tabtitle = "Trade Goods",
}

TA.GUI.station_buy = iup.stationsublist {
    dropdown = "YES",
    action = function()
        TA.GUI.stations_tbl = {}
        local buy_id = sorted.stations[tonumber(TA.GUI.station_buy.value)][2]
        local sell_id = sorted.stations[tonumber(TA.GUI.station_sell.value)][2]
        for _, v in pairs(TA.goods) do
            if (v[buy_id] and v[buy_id].s) and
                (v[sell_id] and v[sell_id].b) and
                (v[sell_id].b > v[buy_id].s) and not
                (not is_premium and v[buy_id].s2 > (v[buy_id].b2 or 0)) then
                table.insert(TA.GUI.stations_tbl, 1, {
                    name = get_goods_color(v.name, v.type) .. v.name,
                    sell = v[sell_id].b,
                    buy = v[buy_id].s,
                    profit_cu = (v[sell_id].b - v[buy_id].s) / (v.vol or 1)
                }
                )
            end
        end
        if #TA.GUI.stations_tbl == 0 then
            table.insert(TA.GUI.stations_tbl, 1, {
                name = "No results found",
                buy = "",
                sell = "",
            })
        end
        TA.GUI.station_matrix:reset(TA.GUI.stations_tbl)
    end
}
TA.GUI.station_sell = iup.stationsublist { dropdown = "YES", action = TA.GUI.station_buy.action }

local stations_sorttbl = {
    [1] = {
        title = "Item Name",
        alignment = "ALEFT",
        fn = function(a, b) return a.name < b.name end,
    },
    [2] = {
        title = "Buy Price",
        alignment = "ARIGHT",
        fn = function(a, b)
            if a.buy == b.buy then return a.name < b.name
            else return sortfunc(a.buy, b.buy)
            end
        end,
    },
    [3] = {
        title = "Sell Price",
        alignment = "ARIGHT",
        fn = function(a, b)
            if a.sell == b.sell then return a.name < b.name
            else return not sortfunc(a.sell, b.sell)
            end
        end,
    },
    [4] = {
        title = "Profit / cu",
        alignment = "ARIGHT",
        fn = function(a, b)
            if a.profit_cu == b.profit_cu then return a.name < b.name
            else return not sortfunc(a.profit_cu, b.profit_cu)
            end
        end,
    },
    update_entry = function(matrix, index, data)
        matrix:setcell(index, 1, " " .. data.name)
        matrix:setcell(index, 2, " " .. price_string(data.buy))
        matrix:setcell(index, 3, " " .. price_string(data.sell))
        matrix:setcell(index, 4, " " .. (data.profit_cu and format("%.2fc", data.profit_cu) or ""))
    end,
}

TA.GUI.station_matrix = create_matrix(stations_sorttbl, 4)

local function selectCurrentStation()
    local currentStation = GetCurrentStationID()
    if not currentStation then
        local sectorid = GetCurrentSectorid()
        for stationid, stationdata in pairs(TA.stations) do
            if stationdata.sectorid == sectorid then
                currentStation = stationid
                break
            end
        end
    end
    if currentStation then
        for i, stationData in pairs(sorted.stations) do
            if currentStation == stationData[2] then
                TA.GUI.station_buy.value = i
                TA.GUI.station_sell.value = i
                break
            end
        end
    end
end

local function selectDestinationStation()
    local destinationSector = NavRoute.GetFinalDestination()
    if destinationSector then
        local sectorid = GetCurrentSectorid()
        if destinationSector == sectorid then
            destinationSector = NavRoute.GetNextHop()
        end
        for i, stationData in pairs(sorted.stations) do
            if destinationSector == TA.stations[stationData[2]].sectorid
            then
                TA.GUI.station_sell.value = i
                break
            end
        end
    end
end

local stations_tab = iup.vbox {
    iup.hbox {
        iup.fill { size = 4 },
        iup.label { title = "Buy: " }, TA.GUI.station_buy,
        iup.fill { size = 12 },
        iup.button { title = "", size = Font.Default .. "x" .. Font.Default,
            image = plugin_path .. "arrows.png",
            bgcolor = "255 255 255 255 *",
            action = function()
                local tmp = TA.GUI.station_buy.value
                TA.GUI.station_buy.value = TA.GUI.station_sell.value
                TA.GUI.station_sell.value = tmp
                TA.GUI.station_buy:action()
            end },
        iup.fill { size = 12 },
        iup.label { title = "Sell: " }, TA.GUI.station_sell,
    },
    TA.GUI.station_matrix,
    tabtitle = "Stations",
    OnShow = function()
        selectDestinationStation()
        TA.GUI.station_buy:action()
    end,
}

local function reset_data()
    TA.GUI.goods_list[1] = nil
    TA.GUI.goods_list_typeids = {}
    TA.GUI.station_matrix:reset({})
    TA.GUI.goods_matrix:reset({})

    local tbl = {}
    for k, v in pairs(TA.goods) do
        table.insert(tbl, { k, get_goods_color(v.name, v.type) .. v.name })
    end
    table.sort(tbl, function(a, b) return a[2] < b[2] end)
    for i, v in ipairs(tbl) do
        TA.GUI.goods_list[i] = v[2]
        TA.GUI.goods_list_typeids[i] = v[1]
    end

    sorted.stations = {}
    for k in pairs(TA.stations) do
        local location = location_string(k)
        table.insert(sorted.stations, { location, k })
    end
    table.sort(sorted.stations, function(a, b) return sortfunc(a[1], b[1]) end)
    TA.GUI.station_buy[1] = nil
    TA.GUI.station_sell[1] = nil

    for i, v in ipairs(sorted.stations) do
        TA.GUI.station_buy[i] = v[1]
        TA.GUI.station_sell[i] = v[1]
    end

    selectCurrentStation()
    selectDestinationStation()

    -- width of the matrix columns
    local width = getwidth(TA.GUI.goods_matrix) - Font.Default
    TA.GUI.goods_matrix.width1 = width / 1.5
    TA.GUI.goods_matrix.width2 = width / 6
    TA.GUI.goods_matrix.width3 = width / 6

    TA.GUI.station_matrix.width1 = width / 2 -- item name
    TA.GUI.station_matrix.width2 = width / 6 -- buy price
    TA.GUI.station_matrix.width3 = width / 6 -- sell price
    TA.GUI.station_matrix.width4 = width / 6 -- profit / cu

    -- iup sucks, better do the dropdown sizing ourselves:
    local dropdown_width = (TA.GUI.MAIN.w - (stations_tab[1][2].w +
        stations_tab[1][7].w + stations_tab[1][5].w + 52)) / 2
    TA.GUI.station_buy.size = dropdown_width
    TA.GUI.station_sell.size = dropdown_width
    iup.Refresh(TA.GUI.MAIN)

    -- populate both tabs
    TA.GUI.goods_list:action(nil, 1, 1)
    TA.GUI.station_buy:action()
end

local function init()
    local load = unspickle(LoadSystemNotes(notesid)) or {}
    if next(load) == nil then
        -- try to load 1.13 database if we don't have a current one
        local charid = GetCharacterID()
        if charid then
            load = unspickle(LoadSystemNotes(charid .. "201")) or {}
        end
        if next(load) ~= nil then
            TA.Print("Importing old format database")
        end
    else
        TA.Print("Loading database")
    end
    TA.stations = load.stations or {}
    if load.version then
        TA.goods = load.goods or {}
        TA.goods_old = load.goods_old or {}
    else -- old format db
        TA.goods = {}
        TA.goods_old = load.goods or {}
    end
    has_old_goods = next(TA.goods_old) ~= nil
    is_premium = IsPremiumSubscription()
    init_done = true
end

function TA.GUI.Show()
    if not init_done then init() end
    reset_data()
    ShowDialog(TA.GUI.MAIN, iup.CENTER, iup.CENTER)
end

TA.GUI.Close = iup.stationbutton { title = "Close", expand = "HORIZONTAL", action = TA.GUI.Hide }
TA.GUI.MAIN = iup.dialog {
    iup.pdasubframebg {
        iup.vbox {
            iup.pda_root_tabs {
                goods_tab,
                stations_tab,
                tabchange_cb = function(_, newtab)
                    if newtab.OnShow then newtab:OnShow() end
                end,
            },
            iup.hbox { TA.GUI.Close },
        },
    },
    topmost = "YES",
    size = tostring(44 * Font.Default) .. "x%90",
    defaultesc = TA.GUI.Close,
    title = TA.Title,
}
TA.GUI.MAIN:map()

--- new database format conversion function
local function convert_database_item(name, typeid)
    if has_old_goods then
        for stationid in pairs(TA.stations) do
            TA.stations[stationid].name = nil -- while we're already here...
            if TA.goods_old[name] and TA.goods_old[name][stationid] then
                if not TA.goods[typeid][stationid] then
                    TA.goods[typeid][stationid] = {
                        b = TA.goods_old[name][stationid].buy or nil,
                        s = TA.goods_old[name][stationid].sell or nil
                    }
                end
                TA.goods_old[name][stationid] = nil
                local has_station = false
                for _, v in pairs(TA.goods_old[name]) do
                    if type(v) == "table" then has_station = true end
                end
                if not has_station then TA.goods_old[name] = nil end
            end
        end
        has_old_goods = next(TA.goods_old) ~= nil
    end
end

local function update_price(typeid, stationid, buy)
    if TA.goods[typeid] and TA.stations[stationid] then
        TA.goods[typeid][stationid] = TA.goods[typeid][stationid] or {}
        TA.goods[typeid][stationid].b = buy
    end
    save_data()
end

-- wrapper for GetBestPriceInfoOfItem(),
-- returns lprice, lstationid, gprice, gstationid or nil
local function getBestPriceOfItem(itemid)
    local lprice, lloc, gprice, gloc = GetBestPriceInfoOfItem(itemid)
    if not (lprice and lloc and gprice and gloc) then return end
    local lsectorid = tonumber(sub(lloc, 1, -3))
    local lstationid = lsectorid * 256 + tonumber(sub(lloc, -2))
    local gsectorid = tonumber(sub(gloc, 1, -3))
    local gstationid = gsectorid * 256 + tonumber(sub(gloc, -2))
    return lprice, lstationid, gprice, gstationid
end

-- updates and saves system and global best price for supplied item
local function updateBestPricesForItem(itemid, typeid, currentStationId)
    currentStationId = currentStationId or GetCurrentStationID()
    if itemid and typeid and itemid ~= typeid then
        local lprice, lstationid, gprice, gstationid = getBestPriceOfItem(itemid)
        local qty = GetInventoryItemQuantity(itemid)
        -- XXX: Any chance of splitting a single unit item off to get the unit
        -- price for consistency with merch prices?
        if lprice and lstationid and (lstationid ~= currentStationId) then
            lprice = floor(lprice / qty)
            update_price(typeid, lstationid, lprice)
        end
        if gprice and gstationid and (gstationid ~= currentStationId) then
            gprice = floor(gprice / qty)
            update_price(typeid, gstationid, gprice)
        end
    end
end

function TA.GUI.MAIN:OnEvent(event, ...)
    if not init_done then init() end
    if event == "SYSTEM_BEST_PRICE" then
        local arg = { ... } -- itemid, price, locationid
        if not arg[2] then return end -- no system best price in unknown system
        local itemid = arg[1]
        local typeid = GetInventoryItemType(itemid)
        if TA.goods[typeid] then -- can only update if we have seen the item
            -- now we don't actually use SYSTEM_BEST_PRICE directly because it
            -- does not provide the galactic best price:
            updateBestPricesForItem(itemid, typeid)
        end
    elseif event == "PLAYER_LOGGED_OUT" or event == "UNLOAD_INTERFACE" then
        if timer:IsActive() then
            timer:Kill()
            save_data_now()
        end
    elseif (event == "ENTERED_STATION" or
        event == "STATION_UPDATE_PRICE") and not GetStationType()
        and GetCurrentStationID() then
        local stationid = GetCurrentStationID()
        if shipMassBugWorkaround then activeShipMass = GetActiveShipMass() end

        -- add station, if new
        if not TA.stations[stationid] then
            TA.Print("Adding current station to database.")
            TA.stations[stationid] = {
                sectorid = GetCurrentSectorid(),
                systemid = GetCurrentSystemid()
            }
        end
        -- update station dynamic economy status
        TA.stations[stationid].dynamic = IsDynamicEconomyStation()

        -- add/update items sold by the station
        local station_merch = {}
        for i = 1, GetNumStationMerch() do
            local info = GetStationMerchInfo(i)
            if info and info.type then
                station_merch[info.itemtype] = true
                TA.goods[info.itemtype] = TA.goods[info.itemtype] or {}
                TA.goods[info.itemtype][stationid] = TA.goods[info.itemtype][stationid] or {}
                TA.goods[info.itemtype][stationid].s = info.price
                TA.goods[info.itemtype][stationid].s2 = info.price and info.price2
                TA.goods[info.itemtype].name = info.name
                TA.goods[info.itemtype].mass = floor(1000 * (info.mass -
                    (info.type == "ship" and activeShipMass or 0)) + 0.5)
                TA.goods[info.itemtype].vol = info.volume
                TA.goods[info.itemtype].type = info.type
                convert_database_item(info.name, info.itemtype)
                -- station merch is always in the player.invbestprices table,
                -- so let's get them, too:
                updateBestPricesForItem(info.itemid, info.itemtype, stationid)
            end
        end

        -- add/update items in station and ship inventory
        for _, info in StationSellableInventoryPairs() do
            -- check if we got a type, otherwise it's not sellable, e.g. ship engines
            if info.type then
                TA.goods[info.itemtype] = TA.goods[info.itemtype] or {}
                TA.goods[info.itemtype][stationid] = TA.goods[info.itemtype][stationid] or {}
                TA.goods[info.itemtype][stationid].b = info.price
                TA.goods[info.itemtype][stationid].b2 = info.price and info.price2
                TA.goods[info.itemtype].name = info.name
                TA.goods[info.itemtype].mass = floor(1000 *
                    (InvManager.GetInventoryItemPrototype(info.itemtype):GetMass() -
                        (info.type == "ship" and activeShipMass or 0)) + 0.5)
                TA.goods[info.itemtype].vol = TA.goods[info.itemtype].vol or
                    GetInventoryItemVolume(info.itemid)
                TA.goods[info.itemtype].type = info.type
                -- best price is not available for station inventory, and for
                -- ship inventory we already get via SYSTEM_BEST_PRICE above
                --updateBestPricesForItem(v.itemid, v.itemtype, stationid)
            end
        end

        -- purge old sell prices for items no longer for sale at this station
        for typeid in pairs(TA.goods) do
            if TA.goods[typeid][stationid] and not station_merch[typeid] then
                if TA.goods[typeid][stationid].b then
                    TA.goods[typeid][stationid].s = nil
                    TA.goods[typeid][stationid].s2 = nil
                else
                    TA.goods[typeid][stationid] = nil
                end
            end
        end

        if event == "ENTERED_STATION" then
            -- don't delay saving after docking
            save_data_now()
        else
            save_data()
        end
    end
end

function TA.GUI.MAIN:k_any(char)
    local _command = gkinterface.GetCommandForKeyboardBind(char)
    if _command == command then
        TA.GUI.Hide()
        return iup.IGNORE
    end
    return iup.CONTINUE
end

local function command_handler(_, args)
    if args then
        if args[1]:lower() == "version" then
            TA.Print(TA.Title)
        elseif args[1]:lower() == "purge" then
            -- TODO: This should have a confirmation
            TA.goods = {}
            TA.goods_old = {}
            TA.stations = {}
            save_data()
            reset_data()
            TA.Print("All data purged.")
        end
    else
        TA.GUI.Show()
    end
end

RegisterUserCommand(command, command_handler)
RegisterEvent(TA.GUI.MAIN, "SYSTEM_BEST_PRICE")
RegisterEvent(TA.GUI.MAIN, "ENTERED_STATION")
RegisterEvent(TA.GUI.MAIN, "STATION_UPDATE_PRICE")
RegisterEvent(TA.GUI.MAIN, "PLAYER_LOGGED_OUT")
RegisterEvent(TA.GUI.MAIN, "UNLOAD_INTERFACE")
